{ Algorithm: Pattern Matching: Regular Expression }

{
	note: This program supports the following operators
		X	: a charactor (ifself)
		XY	: Y appears after X (concatenation)
		X|Y	: matches X or Y (union)
		X*	: 0 or more times of X
		X+	: 1 or more times of X
		(X)	: treat X as a primary (parentheses; closure)
}

program regularExpression(input,output,script);
const maxlen = 30;
	maxline = 1024;
	maxchar = 255;
	maxstate = 128;
	nullchar = chr(255);
	bDebug = true;
type word = packed array[0..maxlen] of char;
	line = packed array[0..maxline] of char;
	tokenType = (charactor,union,lparent,rparent,star,plus,lbrack,rbrack,endmark);
	nodeType = (charOp,catOp,unionOp,closureOp,emptyOp);
	pnode = ^node;
	node = record
		case op : nodeType of
			charOp : (c : char);
			catOp,unionOp : (left,right : pnode);
			closureOp : (link : pnode);
			emptyOp : ();
	end;
	pshift = ^shift;
	shift = record
		next : pshift;
		cond : char;
		dest : integer;
	end;
	automaton = record
		states : array[1..maxstate] of pshift;
		entry, exit : integer;
		count : integer;
	end;
	stateset = set of 1..maxstate;
	dstate = record
		nstates : stateset;
		fin,stop : Boolean;
		shiftlist : pshift;
	end;
	dautomaton = record
		states : array[1..maxstate] of dstate;
		entry : integer;
		count : integer;
	end;
	pshiftset = ^shiftset;
	shiftset = record
		next : pshiftset;
		cond : char;
		dests : stateset;
	end;
var script : text;
	w : word;
	tree : pnode;
	nfa : automaton;
	dfa : dautomaton;
	i, j : integer;




{======= Parser Section =======}

function parse(var pattern : word) : pnode;
label 0;
var curToken : tokenType;
	curChar : char;
	curPos : integer;
	tokenMap : array[0..maxchar] of tokenType;
	i,j,k : integer;
	result : pnode;
	
procedure error(msg : packed array[low..high:integer] of char);
begin
	writeLn('--- syntax error: ', msg);
	result := nil;
	goto 0;
end;

procedure nextToken;
begin
	curChar := pattern[curPos];
	if (maxlen <= curPos) or (chr(0) = curChar) then begin
		curToken := endmark;
	end else begin
		curPos := curPos + 1;
		curToken := tokenMap[curChar];
	end;
end;	{ nextToken }

function readExpression : pnode;
var xe,ye,ze : pnode;

function readTerm : pnode;
var xt,yt,zt : pnode;

function readFactor : pnode;
var xf,yf,zf : pnode;

function readPrimary : pnode;
var xp : pnode;
begin
	if charactor = curToken then begin
		new(xp,charOp);
		xp^.op := charOp;
		xp^.c := curChar;
		nextToken;
	end else if lparent = curToken then begin
		nextToken;
		xp := readExpression;
		if rparent <> curToken then begin
			error('parentheses not closed');
		end else begin
			nextToken;
		end;
	end else begin
		error('not normal charactor');
	end;
	readPrimary := xp;
end;

{ readFactor }
begin
	xf := readPrimary;
	if nil <> xf then begin
		if (star = curToken) or (plus = curToken) then begin
			new(zf,closureOp);
			zf^.op := closureOp;
			zf^.link := xf;
			if plus = curToken then begin
				yf := zf;
				new(zf,catOp);
				zf^.op := catOp;
				zf^.left := xf;
				zf^.right := yf;
			end;
			xf := zf;
			nextToken;
		end;
	end;
	readFactor := xf;
end;

{ readTerm }
begin
	if (union = curToken) or (rparent = curToken) or (endmark = curToken) then begin
		new(xt,emptyOp);
		xt^.op := emptyOp;
	end else begin
		xt := readFactor;
		while (nil <> xt) 
			and (union <> curToken) 
			and (rparent <> curToken) 
			and (endmark <> curToken) do begin
			yt := xt;
			xt := readFactor;
			if nil <> xt then begin
				new(zt,catOp);
				zt^.op := catOp;
				zt^.left := yt;
				zt^.right := xt;
				xt := zt;
			end;
		end;
	end;
	readTerm := xt;
end;

{ readExpression }
begin
	xe := readTerm;
	while (xe <> nil) and (union = curToken) do begin
		nextToken;
		ye := xe;
		xe := readTerm;
		if nil <> xe then begin
			new(ze,unionOp);
			ze^.op := unionOp;
			ze^.left := ye;
			ze^.right := xe;
			xe := ze;
		end;
	end;
	readExpression := xe;
end;

procedure init;
begin
	curPos := 0;
	for i := 0 to maxchar do tokenMap[i] := charactor;
	tokenMap['|'] := union;
	tokenMap['('] := lparent;
	tokenMap[')'] := rparent;
	tokenMap['*'] := star;
	tokenMap['+'] := plus;
end;

{ parse }
begin
	init;
	nextToken;
	result := readExpression;
	if endmark <> curToken then
		error('extra charactor at the end');
0:
	parse := result;
end;

procedure printSyntax(x : pnode);
begin
	if nil <> x then begin
		case x^.op of
		charOp: write('''', x^.c, '''');
		catOp: begin
			write('{cat ');
			printSyntax(x^.left);
			write(' ');
			printSyntax(x^.right);
			write('}');
		end;
		unionOp: begin
			write('{union ');
			printSyntax(x^.left);
			write(' ');
			printSyntax(x^.right);
			write('}');
		end;
		closureOp: begin
			write('{* ');
			printSyntax(x^.link);
			write('}');
		end;
		emptyOp: write('nil');
		end;
	end;
end;

procedure printNode(x : pnode);
begin
	printSyntax(x);
	writeLn;
end;






{======= Finite Automaton Creation =======}

function createAutomaton(tree : pnode; var s : automaton) : Boolean;
label 0;
var result : Boolean;
procedure error(msg : packed array[low..high:integer] of char);
begin
	writeLn('--- automaton creation error: ', msg);
	result := false;
	goto 0;
end;

procedure initAutomaton(var ss : automaton);
begin
	ss.count := 2;
	ss.entry := 1;
	ss.exit := 2;
	ss.states[1] := nil;
	ss.states[2] := nil;
end;

procedure createAutomatonRec(tree : pnode; var ss : automaton; ent, exi : integer);
var n, m : integer;

procedure addShift(var sst : automaton; entr, exit : integer; ch : char);
var sf : pshift;
begin
	new(sf);
	sf^.cond := ch;
	sf^.dest := exit;
	sf^.next := sst.states[entr];
	sst.states[entr] := sf;
end;

function newState(var sst : automaton) : integer;
begin
	if sst.count >= maxstate then error('too many states');
	sst.count := sst.count + 1;
	sst.states[sst.count] := nil;
	newState := sst.count;
end;

{ createAutomatonRec }
begin
	if nil = tree then begin
		error('bad syntax tree');
	end else begin
		case tree^.op of
		charOp: addShift(ss,ent, exi, tree^.c);
		catOp: begin
			n := newState(ss);
			createAutomatonRec(tree^.left,ss,ent,n);
			createAutomatonRec(tree^.right,ss,n,exi);
		end;
		unionOp: begin
			createAutomatonRec(tree^.left,ss,ent,exi);
			createAutomatonRec(tree^.right,ss,ent,exi);
		end;
		closureOp: begin
			n := newState(ss);
			m := newState(ss);
			addShift(ss,ent,n,nullchar);
			addShift(ss,m,n,nullchar);
			addShift(ss,n,exi,nullchar);
			createAutomatonRec(tree^.link,ss,n,m);
		end;
		emptyOp: addShift(ss,ent,exi,nullchar);
		end;
	end;
end;
		
{ createAutomaton }
begin
	result := true;
	initAutomaton(s);
	if nil <> tree then createAutomatonRec(tree, s, s.entry, s.exit);
0:
	createAutomaton := result;
end;

procedure printAutomaton(var s : automaton);
var i : integer;
	x : pshift;
begin
	for i := 1 to s.count do begin
		x := s.states[i];
		if nil <> x then begin
			write('[', i:3, ']');
			while nil <> x do begin
				if nullchar = x^.cond then
					write('{:', x^.dest, '}')
				else
					write('{', x^.cond, ':', x^.dest, '}');
				x := x^.next;
			end;
			writeLn;
		end else if i = s.exit then begin
			writeLn('[',i:3,']{end}');
		end;
	end;
end;




{======= Automaton Conversion Section =======}

procedure printDAutomaton(var d : dautomaton);
var i : integer;
procedure printDState(var ds : dstate);
var x : pshift;
	i : integer;
begin
	if ds.fin then begin
		if ds.stop then write('{end}');
		x := ds.shiftlist;
		while nil <> x do begin
			write('{', x^.cond, ':', x^.dest, '}');
			x := x^.next;
		end;
		for i := 1 to maxstate do begin
			if i in ds.nstates then write('(',i,')');
		end;
	end else begin
		write('*');
	end;
end;

{ printDAutomaton }
begin
	for i := 1 to d.count do begin
		write('[',i:3,']');
		printDState(d.states[i]);
		writeLn;
	end;
end;
	
function convertToDeterministic(var s : automaton; var d : dautomaton) : Boolean;
label 0;
var entry : stateset;
	n : integer;
	dsf : pshift;
	result : Boolean;
	
procedure error(msg : packed array[low..high:integer] of char);
begin
	writeLn('--- automaton conversion error: ', msg);
	result := false;
	goto 0;
end;

function getDState(var sset : stateset) : integer;
var i,n : integer;

procedure initDState(var ds : dstate);
begin
	ds.nstates := sset;
	ds.fin := false;
	ds.stop := (s.exit in sset);
	ds.shiftlist := nil;
end;

{ getDState }
begin
	n := 0;
	i := 1;
	while i <= d.count do begin
		if d.states[i].nstates = sset then begin
			n := i;
			i := d.count;
		end;
		i := i + 1;
	end;
	if 0 = n then begin
		if d.count < maxstate then begin
			d.count := d.count + 1;
			initDState(d.states[d.count]);
			n := d.count;
		end else begin
			error('too many deterministic states');
		end;
	end;
	getDState := n;
end;

procedure translateForwardings(var sset : stateset);
var i : integer;
procedure translateForwardingsRec(n : integer);
var sf : pshift;
begin
	sset := sset + [n];
	sf := s.states[n];
	while nil <> sf do begin
		if (nullchar = sf^.cond) and not (sf^.dest in sset) then
			translateForwardingsRec(sf^.dest);
		sf := sf^.next;
	end;
end;

{ translateForwardings }
begin
	for i := 1 to s.count do begin
		if i in sset then translateForwardingsRec(i);
	end;
end;

function getNextDState : integer;
var i : integer;
begin
	i := 1;
	while (i <= d.count) and (d.states[i].fin) do i := i + 1;
	if i <= d.count then getNextDState := i else getNextDState := 0;
end;

function getShiftSetList(var ss : stateset) : pshiftset;
var i : integer;
	sf : pshift;
	list,p : pshiftset;
	b : Boolean;
begin
	list := nil;
	for i := 1 to s.count do begin
		if i in ss then begin
			sf := s.states[i];
			while nil <> sf do begin
				if nullchar <> sf^.cond then begin
					p := list;
					b := true;
					while nil <> p do begin
						if p^.cond = sf^.cond then begin
							p^.dests := p^.dests + [sf^.dest];
							b := false;
							p := nil;
						end else begin
							p := p^.next;
						end;
					end;
					if b then begin
						new(p);
						p^.cond := sf^.cond;
						p^.dests := [sf^.dest];
						p^.next := list;
						list := p;
					end;
				end;
				sf := sf^.next;
			end;
		end;
	end;
	getShiftSetList := list;
end;

procedure processDState(var ds : dstate);
var list, item : pshiftset;
	n : integer;
begin
	ds.fin := true;
	list := getShiftSetList(ds.nstates);
	while nil <> list do begin
		item := list;
		list := list^.next;
		translateForwardings(item^.dests);
		
		new(dsf);
		dsf^.next := ds.shiftlist;
		dsf^.cond := item^.cond;
		dsf^.dest := getDState(item^.dests);
		ds.shiftlist := dsf;
		
		dispose(item);
	end;
end;

{ convertToDeterministic }
begin
	result := true;
	entry := [s.entry];
	translateForwardings(entry);
	d.entry := getDState(entry);
	
	n := getNextDState;
	while 0 < n do begin
		processDState(d.states[n]);
		n := getNextDState;
	end;
0:
	convertToDeterministic := result;
end;




{======= Matching Phaze =======}
function match(var key : dautomaton; var buf : line; start : integer) : integer;
var i, j, k, s, len : integer;
	cont : Boolean;

function shift(var ds : dstate; ch : char) : integer;
var item : pshift;
	ret : integer;
begin
	ret := 0;
	item := ds.shiftlist;
	while nil <> item do begin
		if ch = item^.cond then begin
			ret := item^.dest;
			item := nil;
		end else begin
			item := item^.next;
		end;
	end;
	shift := ret;
end;
	
{ match }
begin
	buf[maxline] := chr(0);
	
	cont := true;
	s := key.entry;
	i := 0;
	j := start;
	k := -1;

	if (maxline <= start) or (key.states[s].stop) then cont := false;
	while cont do begin
		if key.states[s].stop then begin
			k := j - i;
			cont := false;
		end else if chr(0) = buf[j] then begin
			cont := false;
		end else begin
			s := shift(key.states[s], buf[j]);
			if 0 < s then begin
				i := i + 1;
				j := j + 1;
			end else begin
				j := 1 + j - i;
				i := 0;
				s := key.entry;
			end;
		end;
	end;
	match := k;
end;




{======= Convensional Routines =======}

{ Read Text }
procedure matchText(var ky : dautomaton);
var buf : line;
	i,m,n,rows,count : integer;
begin
	reset(script);
	count := 0;
	rows := 0;
	while not eof(script) do begin
		readLn(script, buf);
		rows := rows + 1;
		m := 1;
		n := match(ky, buf, 0);
		if 0 <= n then begin
			writeLn('Found in line ', rows, ' (hit return to continue):');
			writeLn(buf);
			repeat
				count := count + 1;
				for i := m to n do write(' ');
				write('^');
				m := n + 2;
				n := match(ky, buf, n + 1);
			until 0 > n;
			writeLn;
			readLn;
		end;
	end;
	writeLn('Total ', count, ' times found');
end;

{ Main Routine }
begin
	write('Pattern: ');
	readLn(w);

	tree := parse(w);
	if nil <> tree then begin
		if bDebug then begin
			writeLn('syntax tree:');
			printNode(tree);
			readLn;
		end;
		if createAutomaton(tree, nfa) then begin
			if bDebug then begin
				writeLn('NFA:');
				printAutomaton(nfa);
				readLn;
			end;
			if convertToDeterministic(nfa,dfa) then begin
				if bDebug then begin
					writeLn('DFA:');
					printDAutomaton(dfa);
					readLn;
				end;
				matchText(dfa);
				writeLn('program ended');
			end;
		end;
	end;
end.
