{ Algorithm: Pattern Matching: Regular Expression: Backward Comparison }

{
	note: This program supports the following operators
		X	: a charactor (ifself)
		XY	: Y appears after X (concatenation)
		X|Y	: matches X or Y (union)
		X*	: 0 or more times of X
		X+	: 1 or more times of X
		(X)	: treat X as a primary (parentheses; closure)
		[abc]	: equivalent to (a|b|c)
		[a-c]	: equivalent to (a|b|c)
		[^ab]	: matches any letters other than [ab]
		$	: matches all unprintable letters (delimiters)
		.	: matches all letters
		
		* A line-head and a line-tail is translated into chr(0)
}

program regularExpression3(input,output,script);
const maxlen = 30;
	maxline = 1024;
	maxchar = 255;
	maxstate = 128;
	nullchar = chr(255);
	bDebug = true;
type word = packed array[0..maxlen] of char;
	line = packed array[0..maxline] of char;
	charset = set of chr(0)..chr(maxchar);
	tokenType = (charactor,union,lparent,rparent,star,plus,lbrack,rbrack,dollar,period,endmark);
	nodeType = (charOp,setOp,catOp,unionOp,closureOp,emptyOp);
	pnode = ^node;
	node = record
		case op : nodeType of
			charOp : (c : char);
			setOp : (s : charset);
			catOp,unionOp : (left,right : pnode);
			closureOp : (link : pnode);
			emptyOp : ();
	end;
	pshift = ^shift;
	shift = record
		next : pshift;
		cond : char;
		dest : integer;
	end;
	automaton = record
		map : array[0..maxchar] of char;
		states : array[1..maxstate] of pshift;
		entry, exit : integer;
		count : integer;
	end;
	stateset = set of 1..maxstate;
	dstate = record
		nstates : stateset;
		fin,stop : Boolean;
		minstep : integer;
		shiftlist : pshift;
	end;
	dautomaton = record
		skip : array[0..maxchar] of integer;
		map : array[0..maxchar] of char;
		states : array[1..maxstate] of dstate;
		entry : integer;
		count : integer;
	end;
	pshiftset = ^shiftset;
	shiftset = record
		next : pshiftset;
		cond : char;
		dests : stateset;
	end;
var script : text;
	w : word;
	roottree : pnode;
	nfa : automaton;
	dfa : dautomaton;
	i, j : integer;


{======= Utility Section =======}

function isPrint(ch : char) : Boolean;
begin
	isPrint := ((chr(0x20) <= ch) and (ch <= chr(0x5f))) 
		or ((chr(0x61) <= ch) and (ch <= chr(0x7e)));
end;

procedure printSet(var cs : charset);
var ch : char;
begin
	for ch := chr(0) to chr(maxchar) do begin
		if 0 = ord(ch) mod 16 then writeLn;
		if ch in cs then write('****') else write('----');
	end;
	writeLn;
end;

{======= Parser Section =======}

function parse(var pattern : word) : pnode;
label 0;
var curToken : tokenType;
	curChar : char;
	curPos : integer;
	tokenMap : array[0..maxchar] of tokenType;
	i,j,k : integer;
	result : pnode;
	
procedure error(msg : packed array[low..high:integer] of char);
begin
	writeLn('--- syntax error: ', msg);
	result := nil;
	goto 0;
end;

procedure nextToken;
begin
	curChar := pattern[curPos];
	if (maxlen <= curPos) or (chr(0) = curChar) then begin
		curToken := endmark;
	end else begin
		curPos := curPos + 1;
		if ('\\' = curChar) and (maxlen > curPos) then begin
			curChar := pattern[curPos];
			curPos := curPos + 1;
			curToken := charactor;
		end else begin
			curToken := tokenMap[curChar];
		end;
	end;
end;	{ nextToken }

function readExpression : pnode;
var xe,ye,ze : pnode;

function readTerm : pnode;
var xt,yt,zt : pnode;

function readFactor : pnode;
var xf,yf,zf : pnode;

function readPrimary : pnode;
var xp : pnode;

procedure readCharSet(var cs : charset);
var invert : Boolean;
	ch, prev : char;
begin
	cs := [];
	invert := false;
	if '^' = curChar then begin
		invert := true;
		nextToken;
	end;
	if ']' = curChar then begin
		cs := cs + [']'];
		nextToken;
	end;
	prev := nullchar;
	while (']' <> curChar) and (endmark <> curToken) do begin
		if ('-' = curChar) and (nullchar <> prev) then begin
			nextToken;
			if (']' = curChar) or (endmark = curToken) then begin
				cs := cs + ['-'];
			end else begin
				for ch := prev to curChar do cs := cs + [ch];
				nextToken;
			end;
			prev := nullchar;
		end else begin
			cs := cs + [curChar];
			prev := curChar;
			nextToken;
		end;
	end;

	if rbrack <> curToken then begin
		error('brackets not closed');
	end else begin
		nextToken;
	end;
	
	if invert then cs := - cs;
	if nullchar in cs then cs := cs - [nullchar];
end;

{ readPrimary }
begin
	if charactor = curToken then begin
		new(xp,charOp);
		xp^.op := charOp;
		xp^.c := curChar;
		nextToken;
	end else if lparent = curToken then begin
		nextToken;
		xp := readExpression;
		if rparent <> curToken then begin
			error('parentheses not closed');
		end else begin
			nextToken;
		end;
	end else if lbrack = curToken then begin
		nextToken;
		new(xp,setOp);
		xp^.op := setOp;
		readCharSet(xp^.s);
	end else if period = curToken then begin
		nextToken;
		new(xp,setOp);
		xp^.op := setOp;
		xp^.s := [chr(1)..chr(maxchar)];
	end else if dollar = curToken then begin
		nextToken;
		new(xp,setOp);
		xp^.op := setOp;
		xp^.s := [chr(0)..chr(0x20),chr(0x60),chr(0x7f)..chr(maxchar)];
	end else begin
		error('not normal charactor');
	end;
	readPrimary := xp;
end;

{ readFactor }
begin
	xf := readPrimary;
	if nil <> xf then begin
		if (star = curToken) or (plus = curToken) then begin
			new(zf,closureOp);
			zf^.op := closureOp;
			zf^.link := xf;
			if plus = curToken then begin
				yf := zf;
				new(zf,catOp);
				zf^.op := catOp;
				zf^.left := xf;
				zf^.right := yf;
			end;
			xf := zf;
			nextToken;
		end;
	end;
	readFactor := xf;
end;

{ readTerm }
begin
	if (union = curToken) or (rparent = curToken) or (endmark = curToken) then begin
		new(xt,emptyOp);
		xt^.op := emptyOp;
	end else begin
		xt := readFactor;
		while (nil <> xt) 
			and (union <> curToken) 
			and (rparent <> curToken) 
			and (endmark <> curToken) do begin
			yt := xt;
			xt := readFactor;
			if nil <> xt then begin
				new(zt,catOp);
				zt^.op := catOp;
				zt^.left := yt;
				zt^.right := xt;
				xt := zt;
			end;
		end;
	end;
	readTerm := xt;
end;

{ readExpression }
begin
	xe := readTerm;
	while (xe <> nil) and (union = curToken) do begin
		nextToken;
		ye := xe;
		xe := readTerm;
		if nil <> xe then begin
			new(ze,unionOp);
			ze^.op := unionOp;
			ze^.left := ye;
			ze^.right := xe;
			xe := ze;
		end;
	end;
	readExpression := xe;
end;

procedure init;
begin
	curPos := 0;
	for i := 0 to maxchar do tokenMap[i] := charactor;
	tokenMap['|'] := union;
	tokenMap['('] := lparent;
	tokenMap[')'] := rparent;
	tokenMap['*'] := star;
	tokenMap['+'] := plus;
	tokenMap['['] := lbrack;
	tokenMap[']'] := rbrack;
	tokenMap['.'] := period;
	tokenMap['$'] := dollar;
end;

{ parse }
begin
	init;
	nextToken;
	result := readExpression;
	if endmark <> curToken then
		error('extra charactor at the end');
0:
	parse := result;
end;

procedure printSyntax(x : pnode);
var ci : char;
begin
	if nil <> x then begin
		case x^.op of
		charOp: write('''', x^.c, '''');
		setOp: begin
			write('[');
			for ci := chr(0) to chr(maxchar) do if isPrint(ci) and (ci in x^.s) then write(ci);
			write(']');
		end;
		catOp: begin
			write('{cat ');
			printSyntax(x^.left);
			write(' ');
			printSyntax(x^.right);
			write('}');
		end;
		unionOp: begin
			write('{union ');
			printSyntax(x^.left);
			write(' ');
			printSyntax(x^.right);
			write('}');
		end;
		closureOp: begin
			write('{* ');
			printSyntax(x^.link);
			write('}');
		end;
		emptyOp: write('nil');
		end;
	end;
end;

procedure printNode(x : pnode);
begin
	printSyntax(x);
	writeLn;
end;






{======= Finite Automaton Creation =======}

procedure printMap(var map : array[low..high : char] of char);
var ci,ch : char;
begin
	for ci := low to high do begin
		if ord(ci) mod 16 = 0 then writeLn;
		ch := map[ci];
		if nullchar = ch then begin
			write('   -');
		end else if isPrint(ch) then begin
			write(' <', ch, '>');
		end else begin
			write(ord(ch):4);
		end;
	end;
	writeLn;
end;

procedure printAutomaton(var s : automaton);
var i : integer;
	x : pshift;
	ch,ci : char;
begin
	for i := 1 to s.count do begin
		x := s.states[i];
		if nil <> x then begin
			write('[', i:3, ']');
			while nil <> x do begin
				if nullchar = x^.cond then begin
					write('{:', x^.dest, '}')
				end else begin
					write('{');
					for ci := chr(0) to chr(maxchar) do begin
						if isPrint(ci) and (x^.cond = s.map[ci]) then write(ci);
					end;
					write(':', x^.dest, '}');
				end;
				x := x^.next;
			end;
			writeLn;
		end else if i = s.exit then begin
			writeLn('[',i:3,']{end}');
		end;
	end;
	write('map:');
	printMap(s.map);
end;

function createAutomaton(tree : pnode; var s : automaton) : Boolean;
label 0;
var result : Boolean;
	ch : char;

procedure error(msg : packed array[low..high:integer] of char);
begin
	writeLn('--- automaton creation error: ', msg);
	result := false;
	goto 0;
end;

procedure createAutomatonRec(subtree : pnode; ent, exi : integer);
var n, m : integer;

procedure addForwarding(entr, exit : integer);
var sf : pshift;
begin
	new(sf);
	sf^.cond := nullchar;
	sf^.dest := exit;
	sf^.next := s.states[entr];
	s.states[entr] := sf;
end;

procedure addShift(entr, exit : integer; cond : char);
var sf : pshift;
begin
	new(sf);
	sf^.cond := cond;
	sf^.dest := exit;
	sf^.next := s.states[entr];
	s.states[entr] := sf;
end;

procedure dupShift(g1,g2 : char);
var i : integer;
	sf : pshift;
begin
	for i := 1 to s.count do begin
		sf := s.states[i];
		while nil <> sf do begin
			if g1 = sf^.cond then addShift(i,sf^.dest,g2);
			sf := sf^.next;
		end;
	end;
end;

procedure addShiftS(entr,exit : integer; ch : char);
var x,ci : char;
begin
	addShift(entr,exit,ch);
	if nullchar = s.map[ch] then begin
		{ ch is first-appearance }
		s.map[ch] := ch;
	end else if ch = s.map[ch] then begin
		{ ch is a top of a group }
		x := nullchar;
		for ci := succ(ch) to chr(maxchar) do begin
			if ch = s.map[ci] then begin
				if nullchar = x then x := ci;
				s.map[ci] := x;
			end;
		end;
		if nullchar <> x then dupShift(ch,x);
	end else begin
		{ ch is a non-top member of a group }
		dupShift(s.map[ch],ch);
		s.map[ch] := ch;
	end;
end;

procedure addShiftM(entr,exit : integer; var cs : charset);
var x,y,z,ci,ch : char;
	csTemp : charset;
begin
	csTemp := cs;
	y := nullchar;
	for ch := chr(0) to chr(maxchar) do if ch in csTemp then begin
		if nullchar = s.map[ch] then begin
			{ ch is first-appearance }
			if nullchar = y then y := ch;
			s.map[ch] := y;
		end else if ch = s.map[ch] then begin
			{ ch is a top of a group }
			csTemp := csTemp - [ch];
			x := nullchar;
			for ci := succ(ch) to chr(maxchar) do begin
				if ch = s.map[ci] then begin
					if ci in csTemp then begin
						csTemp := csTemp - [ci];
					end else begin
						if nullchar = x then x := ci;
						s.map[ci] := x;
					end;
				end;
			end;
			if nullchar <> x then dupShift(ch,x);
			addShift(entr,exit,ch);
		end else begin
			{ ch is a non-top member of a group }
			x := s.map[ch];
			for ci := ch to chr(maxchar) do begin
				if x = s.map[ci] then begin
					if ci in csTemp then begin
						s.map[ci] := ch;
						csTemp := csTemp - [ci];
					end;
				end;
			end;
			dupShift(x,ch);
			addShift(entr,exit,ch);
		end;
	end;
	if nullchar <> y then addShift(entr,exit,y);
end;

function newState : integer;
begin
	if s.count >= maxstate then error('too many states');
	s.count := s.count + 1;
	s.states[s.count] := nil;
	newState := s.count;
end;

{ createAutomatonRec }
begin
	if nil = subtree then begin
		error('bad syntax tree');
	end else begin
		case subtree^.op of
		charOp: addShiftS(ent, exi, subtree^.c);
		setOp: addShiftM(ent,exi,subtree^.s);
		catOp: begin
			n := newState;
			createAutomatonRec(subtree^.right,ent,n);
			createAutomatonRec(subtree^.left,n,exi);
		end;
		unionOp: begin
			createAutomatonRec(subtree^.right,ent,exi);
			createAutomatonRec(subtree^.left,ent,exi);
		end;
		closureOp: begin
			n := newState;
			m := newState;
			addForwarding(ent,n);
			addForwarding(m,n);
			addForwarding(n,exi);
			createAutomatonRec(subtree^.link,n,m);
		end;
		emptyOp: addForwarding(ent,exi);
		end;
	end;
end;
		
{ createAutomaton }
begin
	s.count := 2;
	s.entry := 1;
	s.exit := 2;
	s.states[1] := nil;
	s.states[2] := nil;	
	for ch := chr(0) to chr(maxchar) do s.map[ch] := nullchar;
	
	result := true;
	if nil <> tree then createAutomatonRec(tree, s.entry, s.exit);
0:
	createAutomaton := result;
end;




{======= Automaton Conversion Section =======}

procedure printSkip(var skip : array[low..high : char] of integer);
var ci,ch : char;
begin
	for ci := low to high do begin
		if ord(ci) mod 16 = 0 then writeLn;
		write(skip[ci]:3);
	end;
	writeLn;
end;

procedure printDAutomaton(var d : dautomaton);
var i : integer;
procedure printDState(var ds : dstate);
var x : pshift;
	i : integer;
	ci : char;
begin
	if ds.fin then begin
		if ds.stop then write('{end}');
		x := ds.shiftlist;
		while nil <> x do begin
			if nullchar = x^.cond then begin
				write('{:', x^.dest, '}')
			end else begin
				write('{');
				for ci := chr(0) to chr(maxchar) do begin
					if isPrint(ci) and (x^.cond = d.map[ci]) then write(ci);
				end;
				write(':', x^.dest, '}');
			end;
			x := x^.next;
		end;
		for i := 1 to maxstate do begin
			if i in ds.nstates then write('(',i,')');
		end;
	end else begin
		write('*');
	end;
end;

{ printDAutomaton }
begin
	for i := 1 to d.count do begin
		write('[',i:3,']');
		printDState(d.states[i]);
		writeLn;
	end;
	write('skip values:');
	printSkip(d.skip);
end;

function convertToDeterministic(var s : automaton; var d : dautomaton) : Boolean;
label 0;
var entry : stateset;
	n : integer;
	dsf : pshift;
	result : Boolean;
	
procedure error(msg : packed array[low..high:integer] of char);
begin
	writeLn('--- automaton conversion error: ', msg);
	result := false;
	goto 0;
end;

function getDState(var sset : stateset) : integer;
var i,n : integer;

procedure initDState(var ds : dstate);
begin
	ds.nstates := sset;
	ds.fin := false;
	ds.stop := (s.exit in sset);
	ds.shiftlist := nil;
end;

{ getDState }
begin
	n := 0;
	i := 1;
	while i <= d.count do begin
		if d.states[i].nstates = sset then begin
			n := i;
			i := d.count;
		end;
		i := i + 1;
	end;
	if 0 = n then begin
		if d.count < maxstate then begin
			d.count := d.count + 1;
			initDState(d.states[d.count]);
			n := d.count;
		end else begin
			error('too many deterministic states');
		end;
	end;
	getDState := n;
end;

procedure translateForwardings(var sset : stateset);
var i : integer;
procedure translateForwardingsRec(n : integer);
var sf : pshift;
begin
	sset := sset + [n];
	sf := s.states[n];
	while nil <> sf do begin
		if (nullchar = sf^.cond) and not (sf^.dest in sset) then
			translateForwardingsRec(sf^.dest);
		sf := sf^.next;
	end;
end;

{ translateForwardings }
begin
	for i := 1 to s.count do begin
		if i in sset then translateForwardingsRec(i);
	end;
end;

function getNextDState : integer;
var i : integer;
begin
	i := 1;
	while (i <= d.count) and (d.states[i].fin) do i := i + 1;
	if i <= d.count then getNextDState := i else getNextDState := 0;
end;

function getShiftSetList(var ss : stateset) : pshiftset;
var i : integer;
	sf : pshift;
	list,p : pshiftset;
	b : Boolean;
begin
	list := nil;
	for i := 1 to s.count do begin
		if i in ss then begin
			sf := s.states[i];
			while nil <> sf do begin
				if nullchar <> sf^.cond then begin
					p := list;
					b := true;
					while nil <> p do begin
						if p^.cond = sf^.cond then begin
							p^.dests := p^.dests + [sf^.dest];
							b := false;
							p := nil;
						end else begin
							p := p^.next;
						end;
					end;
					if b then begin
						new(p);
						p^.cond := sf^.cond;
						p^.dests := [sf^.dest];
						p^.next := list;
						list := p;
					end;
				end;
				sf := sf^.next;
			end;
		end;
	end;
	getShiftSetList := list;
end;

procedure processDState(var ds : dstate);
var list, item : pshiftset;
	n : integer;
begin
	ds.fin := true;
	list := getShiftSetList(ds.nstates);
	while nil <> list do begin
		item := list;
		list := list^.next;
		translateForwardings(item^.dests);
		
		new(dsf);
		dsf^.next := ds.shiftlist;
		dsf^.cond := item^.cond;
		dsf^.dest := getDState(item^.dests);
		ds.shiftlist := dsf;
		
		dispose(item);
	end;
end;

procedure calcSkip;
var i,n,m : integer;
	sf : pshift;

procedure calcMinStepRec(var ds : dstate; step : integer);
var item : pshift;
begin
	if step < ds.minstep then begin
		ds.minstep := step;
		step := step + 1;
		item := ds.shiftlist;
		while nil <> item do begin
			calcMinStepRec(d.states[item^.dest], step);
			item := item^.next;
		end;
	end;
end;

{ calcSkip }
begin
	for i := 1 to d.count do d.states[i].minstep := maxlen;
	calcMinStepRec(d.states[d.entry],0);
	
	m := maxlen;
	for i := 0 to maxchar do d.skip[i] := maxlen;
	for i := 1 to d.count do with d.states[i] do begin
		sf := shiftlist;
		n := minstep;
		while nil <> sf do begin
			if n < d.skip[sf^.cond] then d.skip[sf^.cond] := n;
			sf := sf^.next;
		end;
		if stop and (n < m) then m := n;
	end;
	for i := 0 to maxchar do if d.skip[i] = maxlen then d.skip[i] := m;
end;

{ convertToDeterministic }
begin
	result := true;
	entry := [s.entry];
	translateForwardings(entry);

	d.map := s.map;	
	d.count := 0;
	d.entry := getDState(entry);
	
	n := getNextDState;
	while 0 < n do begin
		processDState(d.states[n]);
		n := getNextDState;
	end;
	calcSkip;
0:
	convertToDeterministic := result;
end;




{======= Matching Phaze =======}
function match(var key : dautomaton; var buf : line; start : integer) : integer;
var i, j, k, s, t, len : integer;
	cont, brk : Boolean;
	ch : char;

function shift(var ds : dstate; cd : char) : integer;
var item : pshift;
	ret : integer;
begin
	ret := 0;
	item := ds.shiftlist;
	while nil <> item do begin
		if cd = item^.cond then begin
			ret := item^.dest;
			item := nil;
		end else begin
			item := item^.next;
		end;
	end;
	shift := ret;
end;
	
{ match }
begin
	buf[maxline] := chr(0);
	
	cont := true;
	brk := false;
	s := key.entry;
	i := 0;
	j := start;
	k := -1;

	if (maxline <= start) or (key.states[s].stop) then cont := false;
	while cont do begin
		if key.states[s].stop then begin
			k := j + i;
			cont := false;
		end else begin
			if 0 > j then begin
				ch := chr(0);
			end else begin
				ch := key.map[buf[j]];
			end;
			s := shift(key.states[s], ch);
			if 0 < s then begin
				i := i + 1;
				j := j - 1;
			end else begin
				t := 1 + j + i;
				j := j + key.skip[ch];
				if j < t then j := t;
				i := 0;
				s := key.entry;
				if brk or (0 > j ) or (maxline <= j) then
					cont := false
				else if chr(0) = buf[j] then
					brk := true;
			end;
		end;
	end;
	match := k;
end;




{======= Convensional Routines =======}

{ Read Text }
procedure matchText(var ky : dautomaton);
var buf : line;
	i,m,n,rows,count : integer;
begin
	reset(script);
	count := 0;
	rows := 0;
	while not eof(script) do begin
		readLn(script, buf);
		rows := rows + 1;
		m := 1;
		n := match(ky, buf, 0);
		if 0 <= n then begin
			writeLn('Found in line ', rows, ' (hit return to continue):');
			writeLn(buf);
			repeat
				count := count + 1;
				for i := m to n do write(' ');
				write('^');
				m := n + 2;
				n := match(ky, buf, n + 1);
			until 0 > n;
			writeLn;
			readLn;
		end;
	end;
	writeLn('Total ', count, ' times found');
end;

{ Main Routine }
begin
	write('Pattern: ');
	readLn(w);

	roottree := parse(w);
	if nil <> roottree then begin
		if bDebug then begin
			writeLn('syntax tree:');
			printNode(roottree);
			readLn;
		end;
		if createAutomaton(roottree, nfa) then begin
			if bDebug then begin
				writeLn('NFA:');
				printAutomaton(nfa);
				readLn;
			end;
			if convertToDeterministic(nfa,dfa) then begin
				if bDebug then begin
					writeLn('DFA:');
					printDAutomaton(dfa);
					readLn;
				end;
				matchText(dfa);
				writeLn('program ended');
			end;
		end;
	end;
end.
