{ Algorithm: Search: Balanced Binary Search }

program balancedBinarySearch(output,data);
const maxlen = 20;
	maxline = 1024;
type word = packed array[0..maxlen] of char;
	line = packed array[0..maxline] of char;
	pcell = ^cell;
	cell = record
		balanced : Boolean;
		parent, left, right : pcell;
		value : word;
	end;
	
var data : text;
	header : pcell;
	nothing : pcell;
	w : word;
	tolower : array['A'..'Z'] of char;
	c1, c2 : char;

{ Binary Search Algorithm }
procedure init(var root : pcell);
begin
	{ nothing is always balanced }
	new(nothing);
	nothing^.balanced := true;
	nothing^.parent := nothing;
	nothing^.left := nothing;
	nothing^.right := nothing;
	root := nothing;
end;

function newCell(var value : word) : pcell;
var cellNew : pcell;
begin
	new(cellNew);
	cellNew^.balanced := false;
	cellNew^.parent := nothing;
	cellNew^.left := nothing;
	cellNew^.right := nothing;
	cellNew^.value := value;
	newCell := cellNew;
end;

procedure rotateLeft(var list : pcell; center : pcell);
var rt : pcell;
begin
	{ center and center^.right must not be nothing }
	rt := center^.right;
	
	center^.right := rt^.left;
	if nothing <> rt^.left then rt^.left^.parent := center;
	
	if nothing = center^.parent then
		list := rt
	else begin
		if center = center^.parent^.left then
			center^.parent^.left := rt
		else
			center^.parent^.right := rt;
	end;
	rt^.parent := center^.parent;
	
	rt^.left := center;
	center^.parent := rt;
end;

procedure rotateRight(var list : pcell; center : pcell);
var lt : pcell;
begin
	{ center and center^.left must not be nothing }
	lt := center^.left;
	
	center^.left := lt^.right;
	if nothing <> lt^.right then lt^.right^.parent := center;
	
	if nothing = center^.parent then
		list := lt
	else begin
		if center = center^.parent^.right then
			center^.parent^.right := lt
		else
			center^.parent^.left := lt;
	end;
	lt^.parent := center^.parent;
	
	lt^.right := center;
	center^.parent := lt;
end;

procedure balance(var list : pcell; leaf : pcell);
var me, uncle, grandpa : pcell;
begin
	{ list and leaf must not be nil }
	list^.balanced := true;
	me := leaf;
	while not me^.parent^.balanced do begin
		{ me is neither root nor root's child }
		grandpa := me^.parent^.parent;
		if me^.parent = grandpa^.left then begin
			uncle := grandpa^.right;
			if uncle^.balanced then begin
				if me = me^.parent^.right then begin
					me := me^.parent;
					rotateLeft(list, me);
				end;
				me^.parent^.balanced := true;
				grandpa^.balanced := false;
				rotateRight(list, grandpa);
			end else begin
				me^.parent^.balanced := true;
				uncle^.balanced := true;
				grandpa^.balanced := false;
				me := grandpa;
			end;
		end else begin
			uncle := grandpa^.left;
			if uncle^.balanced then begin
				if me = me^.parent^.left then begin
					me := me^.parent;
					rotateRight(list, me);
				end;
				me^.parent^.balanced := true;
				grandpa^.balanced := false;
				rotateLeft(list, grandpa);
			end else begin
				me^.parent^.balanced := true;
				uncle^.balanced := true;
				grandpa^.balanced := false;
				me := grandpa;
			end;
		end;
		list^.balanced := true;
	end;
end;

function insert(list : pcell; var value : word) : pcell;
label 100;
type linkType = (linkLeft, linkRight);
var cellChild, cellParent : pcell;
	link : linkType;
begin
	cellParent := nothing;
	cellChild := list;
	while nothing <> cellChild do begin
		if cellChild^.value < value then begin
			cellParent := cellChild;
			link := linkRight;
			cellChild := cellChild^.right;
		end else if cellChild^.value > value then begin
			cellParent := cellChild;
			link := linkLeft;
			cellChild := cellChild^.left;
		end else
			goto 100;
	end;
100:
	if nothing = cellChild then begin
		cellChild := newCell(value);
		cellChild^.parent := cellParent;
		if nothing = cellParent then
			list := cellChild
		else if linkLeft = link then
			cellParent^.left := cellChild
		else
			cellParent^.right := cellChild;
		balance(list, cellChild);
	end;
	insert := list;
end;

procedure print(list : pcell);
begin
	if nothing <> list then begin
		print(list^.left);
		writeLn(list^.value);
		print(list^.right);
	end;
end;

{ Text Parsing }
function parseLine(list : pcell; var buf : line) : pcell;
var i, j, k, len : integer;
	value : word;
	ch : char;
begin
	len := 0;
	for i := 0 to maxline do begin
		ch := buf[i];
		if ('A' <= ch) and ('Z' >= ch) then ch := tolower[ch];
		if ('a' <= ch) and ('z' >= ch) then begin
			if len < maxlen then begin
				value[len] := ch;
				len := len + 1;
			end;
		end else begin
			if 2 < len then begin
				for k := len to maxlen do value[k] := chr(0);
				list := insert(list, value);
			end;
			len := 0;
			if chr(0) = ch then break;
		end;
	end;
	parseLine := list;
end;

function parseText(list : pcell) : pcell;
var buf : line;
begin
	reset(data);
	while not eof(data) do begin
		readLn(data, buf);
		writeLn(buf);
		list := parseLine(list, buf);
	end;
	parseText := list;
end;

{ Main Routine }
begin
	c2 := 'a';
	for c1 := 'A' to 'Z' do begin
		tolower[c1] := c2;
		c2 := succ(c2);
	end;

	init(header);
	
	writeLn('Input Text:');
	header := parseText(header);

	writeLn('Contained Words:');
	print(header);
end.
