{ Algorithm: Search: B-Tree Search }

program bTreeSearch(input, output,data);
const maxlen = 20;
	maxline = 1024;
	maxtag = 10;
type word = packed array[0..maxlen] of char;
	line = packed array[0..maxline] of char;
	pcell = ^cell;
	nodeKind = (leaf, subtree);
	cell = record
		case phenotype : nodeKind of
			leaf : (
				value : word;
			);
			subtree : (
				count : integer;
				child : array[0..maxtag] of pcell;
				tag : array[1..maxtag] of word;
			);
	end;
	
var data : text;
	header : pcell;
	w : word;
	tolower : array['A'..'Z'] of char;
	c1, c2 : char;

{ B-Tree Search Algorithm }
procedure init(var root : pcell);
var i : integer;
begin
	root := nil;
end;

function newLeaf(var value : word) : pcell;
var ce : pcell;
begin
	new(ce, leaf);
	ce^.phenotype := leaf;
	ce^.value := value;
	newLeaf := ce;
end;

function newTree(first : pcell) : pcell;
var ce : pcell;
begin
	new(ce, subtree);
	ce^.phenotype := subtree;
	ce^.child[0] := first;
	ce^.count := 0;
	newTree := ce;
end;

function insert_rec(var tree, ins : pcell; var instag, value : word) : pcell;
var me, parent, ce : pcell;
	tag : word;
	n, m, pos : integer;
begin
	ins := nil;
	me := tree;
	if leaf = me^.phenotype then begin
		if me^.value < value then begin
			ce := newLeaf(value);
			ins := ce;
			instag := value;
		end else if value < me^.value then begin
			ce := newLeaf(value);
			tree := ce;
			ins := me;
			instag := me^.value;
		end else begin
			ce := nil;
		end;
	end else begin	
		pos := 0;
		parent := me;
		n := parent^.count;
		while 0 < n do begin
			if parent^.tag[n] <= value then begin
				pos := n;
				n := 0;
			end else
				n := n - 1;
		end;
		ce := insert_rec(parent^.child[pos], me, tag, value);
		if nil <> me then begin
			if maxtag > parent^.count then begin
				pos := pos + 1;
				for n := parent^.count downto pos do begin
					parent^.child[n+1] := parent^.child[n];
					parent^.tag[n+1] := parent^.tag[n];
				end;
				parent^.child[pos] := me;
				parent^.tag[pos] := tag;
				parent^.count := parent^.count + 1;
			end else begin
				if maxtag div 2 > pos then begin
					ce := newTree(parent^.child[maxtag div 2]);
					ins := ce;
					instag := parent^.tag[maxtag div 2];
					m := 1;
					for n := maxtag div 2 + 1 to maxtag do begin
						ce^.child[m] := parent^.child[n];
						ce^.tag[m] := parent^.tag[n];
						m := m + 1;
					end;
					ce^.count := parent^.count - maxtag div 2;
					pos := pos + 1;
					for n := maxtag div 2 - 1 downto pos do begin
						parent^.child[n+1] := parent^.child[n];
						parent^.tag[n+1] := parent^.tag[n];
					end;
					parent^.child[pos] := me;
					parent^.tag[pos] := tag;
					parent^.count := maxtag div 2;
				end else if maxtag div 2 = pos then begin
					ce := newTree(me);
					ins := ce;
					instag := tag;
					m := 1;
					for n := maxtag div 2 + 1 to maxtag do begin
						ce^.child[m] := parent^.child[n];
						ce^.tag[m] := parent^.tag[n];
						m := m + 1;
					end;
					ce^.count := parent^.count - maxtag div 2;
					parent^.count := maxtag div 2;
				end else begin
					ce := newTree(parent^.child[maxtag div 2 + 1]);
					ins := ce;
					instag := parent^.tag[maxtag div 2 + 1];
					m := 1;
					for n := maxtag div 2 + 2 to pos do begin
						ce^.child[m] := parent^.child[n];
						ce^.tag[m] := parent^.tag[n];
						m := m + 1;
					end;
					ce^.child[m] := me;
					ce^.tag[m] := tag;
					m := m + 1;
					for n := pos + 1 to maxtag do begin
						ce^.child[m] := parent^.child[n];
						ce^.tag[m] := parent^.tag[n];
						m := m + 1;
					end;
					ce^.count := parent^.count - maxtag div 2;
					parent^.count := maxtag div 2;
				end;
			end;
		end;
	end;
	insert_rec := ce;
end;

procedure insert(var tree : pcell; var value : word);
var ce, ins : pcell;
	tag : word;
begin
	if nil = tree then begin
		ce := newLeaf(value);
		tree := ce;
	end else begin
		ce := insert_rec(tree, ins, tag, value);
		if nil <> ins then begin
			ce := newTree(tree);
			ce^.count := 1;
			ce^.tag[1] := tag;
			ce^.child[1] := ins;
			tree := ce;
		end;
	end;
end;

procedure print(tree : pcell);
var i : integer;
begin
	if nil <> tree then begin
		if leaf = tree^.phenotype then begin
			writeLn(tree^.value);
		end else begin
			for i := 0 to tree^.count do print(tree^.child[i]);
		end;
	end;
end;

{ Text Parsing }
function parseLine(list : pcell; var buf : line) : pcell;
var i, j, k, len : integer;
	value : word;
	ch : char;
begin
	len := 0;
	for i := 0 to maxline do begin
		ch := buf[i];
		if ('A' <= ch) and ('Z' >= ch) then ch := tolower[ch];
		if ('a' <= ch) and ('z' >= ch) then begin
			if len < maxlen then begin
				value[len] := ch;
				len := len + 1;
			end;
		end else begin
			if 2 < len then begin
				for k := len to maxlen do value[k] := chr(0);
				insert(list, value);
			end;
			len := 0;
			if chr(0) = ch then break;
		end;
	end;
	parseLine := list;
end;

function parseText(list : pcell) : pcell;
var buf : line;
begin
	reset(data);
	while not eof(data) do begin
		readLn(data, buf);
		writeLn(buf);
		list := parseLine(list, buf);
{		print(list);
		readLn;}
	end;
	parseText := list;
end;

{ Main Routine }
begin
	c2 := 'a';
	for c1 := 'A' to 'Z' do begin
		tolower[c1] := c2;
		c2 := succ(c2);
	end;
	
	writeLn('Input Text:');
	header := parseText(header);

	writeLn('Contained Words:');
	print(header);
end.
